import sys
from enum import StrEnum, auto
from importlib import import_module
from pathlib import Path
from typing import Protocol, Self

from pydantic import BaseModel, ConfigDict
from rich import print

from gimad._db import DatabaseClient
from gimad._utils import ONEOFF_DIR, PERMANENT_DIR


class MigrationError(Exception):
    """General migration error"""


class MigrationType(StrEnum):
    PERMANENT = auto()
    ONEOFF = auto()


class MigrationModule(Protocol):
    def up(self) -> None: ...
    def down(self) -> None: ...


class Migration(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    name: str
    type: MigrationType

    @classmethod
    def from_path(cls, p: Path) -> Self:
        return cls.model_validate({"name": p.name, "type": p.parent.name})

    @property
    def module(self) -> MigrationModule:
        m = import_module(self.name.removesuffix(".py"))
        for func in ["up", "down"]:
            if not hasattr(m, func):
                msg = f"Migration file {self.type}/{self.name} missing `{func}()` function"
                raise MigrationError(msg)
        return m  # type: ignore


def _append_paths() -> None:
    sys.path.append(str(ONEOFF_DIR.resolve()))
    sys.path.append(str(PERMANENT_DIR.resolve()))


def _collect_migrations(skip_oneoff: bool = False) -> list[Migration]:
    """Collect all migration scripts from migration script directory"""
    scripts = [Migration.from_path(p) for p in PERMANENT_DIR.glob("*.py")]
    if not skip_oneoff:
        oneoff_scripts = ONEOFF_DIR.glob("*.py")
        scripts.extend(Migration.from_path(p) for p in oneoff_scripts)
    return scripts


class Runner:
    """Migration runner"""

    def __init__(self, client: DatabaseClient) -> None:
        _append_paths()
        self._client = client
        self._client.setup_history_table()

    def up(self, n: int | None = None, skip_oneoff: bool = False) -> None:
        """
        Run the upgrade code path of all pending migrations

        If `n` is specfied, only run the first `n` migrations.
        """
        self._migrations = self._collect_pending_migrations(skip_oneoff)
        if not self._migrations:
            print("[blue]Nothing to run[/blue]")
            return

        if n:
            self._up_with_count(n)
            return

        for migration in self._migrations:
            self._migrate(migration)

    def _up_with_count(self, n: int) -> None:
        for migration in self._migrations:
            if n < 1:
                return
            self._migrate(migration)
            n -= 1

    def down(self, n: int) -> None:
        """Run the downgrade code path of `n` latest migration"""
        self._migrations = self._collect_executed_migrations(n)
        for migration in reversed(self._migrations):
            if n < 1:
                return
            self._rollback(migration)
            n -= 1

    def _collect_pending_migrations(self, skip_oneoff: bool) -> list[Migration]:
        scripts = _collect_migrations(skip_oneoff)
        pending = set(self._client.exclude_executed_scripts([s.name for s in scripts]))
        return sorted((s for s in scripts if s.name in pending), key=lambda s: s.name)

    def _collect_executed_migrations(self, n: int) -> list[Migration]:
        rows = self._client.query_last_executed_migrations(n)
        return [Migration.model_validate(r) for r in rows]

    def _migrate(self, migration: Migration) -> None:
        print(f"[blue]:: Migrating {migration.type}/{migration.name}[/blue]")
        migration.module.up()
        self._client.mark_executed(migration.name, migration.type)

    def _rollback(self, migration: Migration) -> None:
        print(f"[blue]:: Rolling back {migration.type}/{migration.name}[/blue]")
        migration.module.down()
        self._client.unmark_executed(migration.name, migration.type)



from pydantic_settings import (
    BaseSettings,
    JsonConfigSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
    YamlConfigSettingsSource,
)

from gimad._utils import CONFIG_NAME


class Config(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="GIMAD_",
        json_file=f"{CONFIG_NAME}.json",
        yaml_file=f"{CONFIG_NAME}.yaml",
        toml_file=f"{CONFIG_NAME}.toml",
    )

    db_url: str

    @classmethod
    def settings_customise_sources(  # noqa: PLR0913
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            env_settings,
            JsonConfigSettingsSource(settings_cls),
            TomlConfigSettingsSource(settings_cls),
            YamlConfigSettingsSource(settings_cls),
            init_settings,
        )


def load_config() -> Config:
    return Config()  # type:ignore



import contextlib
from collections.abc import Sequence
from typing import Self

import psycopg
from psycopg.errors import DuplicateTable


class DatabaseClient:
    def __init__(self, db_url: str) -> None:
        self.db_url = db_url

    def __enter__(self) -> Self:
        self._connection = psycopg.connect(self.db_url, autocommit=True)
        self._cursor = self._connection.cursor()
        return self

    def __exit__(self, type: object, value: object, traceback: object) -> None:
        self._cursor.close()
        self._connection.close()

    def setup_history_table(self) -> None:
        with contextlib.suppress(DuplicateTable):
            self._cursor.execute(
                """
                create table data_migrations(
                    name varchar primary key,
                    type varchar not null,
                    created_at timestamptz not null default now()
                )
                """,
            )

    def exclude_executed_scripts(self, scripts: Sequence[str]) -> list:
        res = self._cursor.execute(
            """
            with all_scripts as (
                select unnest((%s)::varchar[]) name
                )
            select name
            from all_scripts
            where all_scripts.name not in (select name from data_migrations)
            """,
            [scripts],
        )
        return [r[0] for r in res]

    def query_last_executed_migrations(self, n: int) -> list[dict[str, str]]:
        rows = self._cursor.execute(
            """
            select name, type
            from data_migrations
            order by name desc
            limit %s
            """,
            [n],
        )
        return [{"name": r[0], "type": r[1]} for r in rows]

    def mark_executed(self, name: str, type: str) -> None:
        self._cursor.execute(
            """
            insert into data_migrations(name, type)
            values(%s, %s)
            """,
            [name, type],
        )

    def unmark_executed(self, name: str, type: str) -> None:
        self._cursor.execute(
            """
            delete from data_migrations
            where name = %s and type = %s
            """,
            [name, type],
        )



import os

from typer import Typer

from gimad._commands.down import down
from gimad._commands.init import init
from gimad._commands.new import new
from gimad._commands.redo import redo
from gimad._commands.up import up
from gimad._utils import error

app = Typer(
    no_args_is_help=True,
    help="""
Non-schema migration runner for PostgreSQL

For debugging, set the environment variable DEBUG=1 to see the stack trace
    """,
)
init = app.command()(init)
new = app.command()(new)
up = app.command()(up)
down = app.command()(down)
redo = app.command()(redo)


def main() -> None:
    try:
        app()
    except Exception as e:  # noqa: BLE001
        if not os.environ.get("DEBUG"):
            error(str(e))
        else:
            raise



from typing import Annotated, Optional

from typer import Option

from gimad._config import load_config
from gimad._db import DatabaseClient
from gimad._runner import Runner


def up(
    count: Annotated[
        # Typer limitation. See https://github.com/tiangolo/typer/pull/739
        Optional[int],  # noqa: UP007
        Option("-c", "--count", help="Number of migrations to run"),
    ] = None,
    skip_oneoff: Annotated[
        bool,
        Option(
            help="""
Only run permanent migrations

This option is useful for initializing a new database.
            """,
        ),
    ] = False,
) -> None:
    """Run pending migrations"""
    config = load_config()
    with DatabaseClient(config.db_url) as client:
        Runner(client).up(count, skip_oneoff)
